package config

import (
	"errors"
	"time"

	"go.uber.org/zap/zapcore"

	"github.com/kelseyhightower/envconfig"
)

func New() Config {
	config := Config{}
	envconfig.MustProcess("plasma", &config)

	if config.AccessLog.Out == "" {
		config.AccessLog.Out = "stdout"
	}
	if config.ErrorLog.Out == "" {
		config.ErrorLog.Out = "stderr"
	}
	if config.AccessLog.Level == "" {
		config.AccessLog.Level = "debug"
	}

	if config.ErrorLog.Level == "" {
		config.ErrorLog.Level = "debug"
	}

	return config
}

type Config struct {
	AccessLog   Log `envconfig:"ACCESS_LOG"`
	ErrorLog    Log `envconfig:"ERROR_LOG"`
	Debug       bool
	Origin      string
	Port        string `default:"8080"`
	GrpcPort    string `default:"50051"`
	MerticsPort string `default:"9999"`
	SSE         ServerSentEvent
	Subscriber  Subscriber
	TLS         Cert `envconfig:"TLS"`
	Metrics     Metrics
	Pprof       Pprof
}

type ServerSentEvent struct {
	Retry      int    `default:"2000"`
	EventQuery string `default:"eventType"`
}

type Subscriber struct {
	Type  string `default:"mock"`
	Redis Redis
	Mock  Mock
}

type Mock struct {
	Interval time.Duration `default:"1s"`
}

type Channels []string

func (cs Channels) MarshalLogArray(enc zapcore.ArrayEncoder) error {
	for _, c := range cs {
		enc.AppendString(c)
	}
	return nil
}

type OverMaxRetryBehavior struct {
	Type string
}

func (b *OverMaxRetryBehavior) UnmarshalText(text []byte) error {
	switch string(text) {
	case OverMaxRetryBehaviorAlive:
		b.Type = OverMaxRetryBehaviorAlive
	case OverMaxRetryBehaviorDie:
		b.Type = OverMaxRetryBehaviorDie
	default:
		return errors.New("unknown OverMaxRetryBehavior type: " + string(text))
	}

	return nil
}

const (
	OverMaxRetryBehaviorAlive = "alive"
	OverMaxRetryBehaviorDie   = "die"
)

type Redis struct {
	Addr                 string `default:"localhost:6379"`
	Password             string
	DB                   int
	Channels             Channels
	OverMaxRetryBehavior OverMaxRetryBehavior `envconfig:"OVER_MAX_RETRY_BEHAVIOR"`
	MaxRetry             int                  `default:"5"`
	Timeout              time.Duration        `default:"1s"`
	RetryInterval        time.Duration        `default:"5s" envconfig:"RETRY_INTERVAL"`
}

func (r Redis) MarshalLogObject(enc zapcore.ObjectEncoder) error {
	enc.AddString("Addr", r.Addr)
	enc.AddString("Password", r.Password)
	enc.AddInt("DB", r.DB)
	return enc.AddArray("Channels", r.Channels)
}

type Log struct {
	Out   string
	Level string
}

type Cert struct {
	CertFile string `envconfig:"CERT_FILE"`
	KeyFile  string `envconfig:"KEY_FILE"`
}

type Metrics struct {
	Type     string
	Log      LogMetrics
	Syslog   SyslogMetrics
	Interval time.Duration `default:"10s"`
}

type LogMetrics struct {
	Out      string        `default:"stdout"`
	Prefix   string        `default:"metrics"`
	Flag     int           `default:"4"`
	Interval time.Duration `default:"1m"`
}

type SyslogMetrics struct {
	Tag      string        `default:"plasma"`
	Interval time.Duration `default:"1m"`
	Severity int           `default:"6"`
	Facility int
	Network  string
	Addr     string
}

type Pprof struct {
	Host string `default:"0.0.0.0"`
	Port string `default:"6060"`
}
